@@ -98,7 +98,7 @@ struct wlr_gles2_texture *gles2_get_texture(
 
 void push_gles2_marker(const char *file, const char *func);
 void pop_gles2_marker(void);
-#define PUSH_GLES2_DEBUG push_gles2_marker(_wlr_strip_path(__FILE__), __func__)
+#define PUSH_GLES2_DEBUG push_gles2_marker(_WLR_FILENAME, __func__)
 #define POP_GLES2_DEBUG pop_gles2_marker()
 
 #endif
