@@ -10,11 +10,23 @@
 #define WLR_TYPES_WLR_SEAT_H
 
 #include <time.h>
-#include <wayland-server.h>
+#include <wayland-server-core.h>
 #include <wlr/types/wlr_input_device.h>
 #include <wlr/types/wlr_keyboard.h>
 #include <wlr/types/wlr_surface.h>
 
+#define WLR_SERIAL_RINGSET_SIZE 128
+
+struct wlr_serial_range {
+	uint32_t min_incl;
+	uint32_t max_incl;
+};
+struct wlr_serial_ringset {
+	struct wlr_serial_range data[WLR_SERIAL_RINGSET_SIZE];
+	int end;
+	int count;
+};
+
 /**
  * Contains state for a single client's bound wl_seat resource and can be used
  * to issue input events to that client. The lifetime of these objects is
@@ -35,6 +47,10 @@ struct wlr_seat_client {
 	struct {
 		struct wl_signal destroy;
 	} events;
+
+	// set of serials which were sent to the client on this seat
+	// for use by wlr_seat_client_{next_serial,validate_event_serial}
+	struct wlr_serial_ringset serials;
 };
 
 struct wlr_touch_point {
@@ -61,11 +77,11 @@ struct wlr_seat_pointer_grab;
 struct wlr_pointer_grab_interface {
 	void (*enter)(struct wlr_seat_pointer_grab *grab,
 			struct wlr_surface *surface, double sx, double sy);
-	void (*motion)(struct wlr_seat_pointer_grab *grab, uint32_t time,
+	void (*motion)(struct wlr_seat_pointer_grab *grab, uint32_t time_msec,
 			double sx, double sy);
-	uint32_t (*button)(struct wlr_seat_pointer_grab *grab, uint32_t time,
-			uint32_t button, uint32_t state);
-	void (*axis)(struct wlr_seat_pointer_grab *grab, uint32_t time,
+	uint32_t (*button)(struct wlr_seat_pointer_grab *grab, uint32_t time_msec,
+			uint32_t button, enum wlr_button_state state);
+	void (*axis)(struct wlr_seat_pointer_grab *grab, uint32_t time_msec,
 			enum wlr_axis_orientation orientation, double value,
 			int32_t value_discrete, enum wlr_axis_source source);
 	void (*frame)(struct wlr_seat_pointer_grab *grab);
@@ -78,7 +94,7 @@ struct wlr_keyboard_grab_interface {
 	void (*enter)(struct wlr_seat_keyboard_grab *grab,
 			struct wlr_surface *surface, uint32_t keycodes[],
 			size_t num_keycodes, struct wlr_keyboard_modifiers *modifiers);
-	void (*key)(struct wlr_seat_keyboard_grab *grab, uint32_t time,
+	void (*key)(struct wlr_seat_keyboard_grab *grab, uint32_t time_msec,
 			uint32_t key, uint32_t state);
 	void (*modifiers)(struct wlr_seat_keyboard_grab *grab,
 			struct wlr_keyboard_modifiers *modifiers);
@@ -88,13 +104,13 @@ struct wlr_keyboard_grab_interface {
 struct wlr_seat_touch_grab;
 
 struct wlr_touch_grab_interface {
-	uint32_t (*down)(struct wlr_seat_touch_grab *grab, uint32_t time,
+	uint32_t (*down)(struct wlr_seat_touch_grab *grab, uint32_t time_msec,
 			struct wlr_touch_point *point);
-	void (*up)(struct wlr_seat_touch_grab *grab, uint32_t time,
+	void (*up)(struct wlr_seat_touch_grab *grab, uint32_t time_msec,
 			struct wlr_touch_point *point);
-	void (*motion)(struct wlr_seat_touch_grab *grab, uint32_t time,
+	void (*motion)(struct wlr_seat_touch_grab *grab, uint32_t time_msec,
 			struct wlr_touch_point *point);
-	void (*enter)(struct wlr_seat_touch_grab *grab, uint32_t time,
+	void (*enter)(struct wlr_seat_touch_grab *grab, uint32_t time_msec,
 			struct wlr_touch_point *point);
 	// XXX this will conflict with the actual touch cancel which is different so
 	// we need to rename this
@@ -131,6 +147,8 @@ struct wlr_seat_pointer_grab {
 	void *data;
 };
 
+#define WLR_POINTER_BUTTONS_CAP 16
+
 struct wlr_seat_pointer_state {
 	struct wlr_seat *seat;
 	struct wlr_seat_client *focused_client;
@@ -140,7 +158,8 @@ struct wlr_seat_pointer_state {
 	struct wlr_seat_pointer_grab *grab;
 	struct wlr_seat_pointer_grab *default_grab;
 
-	uint32_t button_count;
+	uint32_t buttons[WLR_POINTER_BUTTONS_CAP];
+	size_t button_count;
 	uint32_t grab_button;
 	uint32_t grab_serial;
 	uint32_t grab_time;
@@ -191,7 +210,6 @@ struct wlr_seat {
 	struct wl_global *global;
 	struct wl_display *display;
 	struct wl_list clients;
-	struct wl_list drag_icons; // wlr_drag_icon::link
 
 	char *name;
 	uint32_t capabilities;
@@ -239,8 +257,9 @@ struct wlr_seat {
 		struct wl_signal request_set_primary_selection;
 		struct wl_signal set_primary_selection;
 
+		// wlr_seat_request_start_drag_event
+		struct wl_signal request_start_drag;
 		struct wl_signal start_drag; // wlr_drag
-		struct wl_signal new_drag_icon; // wlr_drag_icon
 
 		struct wl_signal destroy;
 	} events;
@@ -265,6 +284,12 @@ struct wlr_seat_request_set_primary_selection_event {
 	uint32_t serial;
 };
 
+struct wlr_seat_request_start_drag_event {
+	struct wlr_drag *drag;
+	struct wlr_surface *origin;
+	uint32_t serial;
+};
+
 struct wlr_seat_pointer_focus_change_event {
 	struct wlr_seat *seat;
 	struct wlr_surface *old_surface, *new_surface;
@@ -329,7 +354,7 @@ void wlr_seat_pointer_clear_focus(struct wlr_seat *wlr_seat);
  * `wlr_seat_pointer_notify_motion()` to send motion events to respect pointer
  * grabs.
  */
-void wlr_seat_pointer_send_motion(struct wlr_seat *wlr_seat, uint32_t time,
+void wlr_seat_pointer_send_motion(struct wlr_seat *wlr_seat, uint32_t time_msec,
 		double sx, double sy);
 
 /**
@@ -338,15 +363,15 @@ void wlr_seat_pointer_send_motion(struct wlr_seat *wlr_seat, uint32_t time,
  * `wlr_seat_pointer_notify_button()` to send button events to respect pointer
  * grabs.
  */
-uint32_t wlr_seat_pointer_send_button(struct wlr_seat *wlr_seat, uint32_t time,
-		uint32_t button, uint32_t state);
+uint32_t wlr_seat_pointer_send_button(struct wlr_seat *wlr_seat,
+		uint32_t time_msec, uint32_t button, enum wlr_button_state state);
 
 /**
  * Send an axis event to the surface with pointer focus. Compositors should use
  * `wlr_seat_pointer_notify_axis()` to send axis events to respect pointer
  * grabs.
  **/
-void wlr_seat_pointer_send_axis(struct wlr_seat *wlr_seat, uint32_t time,
+void wlr_seat_pointer_send_axis(struct wlr_seat *wlr_seat, uint32_t time_msec,
 		enum wlr_axis_orientation orientation, double value,
 		int32_t value_discrete, enum wlr_axis_source source);
 
@@ -382,20 +407,20 @@ void wlr_seat_pointer_notify_enter(struct wlr_seat *wlr_seat,
  * Notify the seat of motion over the given surface. Pass surface-local
  * coordinates where the pointer motion occurred.
  */
-void wlr_seat_pointer_notify_motion(struct wlr_seat *wlr_seat, uint32_t time,
-		double sx, double sy);
+void wlr_seat_pointer_notify_motion(struct wlr_seat *wlr_seat,
+		uint32_t time_msec, double sx, double sy);
 
 /**
  * Notify the seat that a button has been pressed. Returns the serial of the
  * button press or zero if no button press was sent.
  */
 uint32_t wlr_seat_pointer_notify_button(struct wlr_seat *wlr_seat,
-		uint32_t time, uint32_t button, uint32_t state);
+		uint32_t time_msec, uint32_t button, enum wlr_button_state state);
 
 /**
  * Notify the seat of an axis event.
  */
-void wlr_seat_pointer_notify_axis(struct wlr_seat *wlr_seat, uint32_t time,
+void wlr_seat_pointer_notify_axis(struct wlr_seat *wlr_seat, uint32_t time_msec,
 		enum wlr_axis_orientation orientation, double value,
 		int32_t value_discrete, enum wlr_axis_source source);
 
@@ -438,14 +463,14 @@ void wlr_seat_keyboard_end_grab(struct wlr_seat *wlr_seat);
  * Send the keyboard key to focused keyboard resources. Compositors should use
  * `wlr_seat_notify_key()` to respect keyboard grabs.
  */
-void wlr_seat_keyboard_send_key(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_keyboard_send_key(struct wlr_seat *seat, uint32_t time_msec,
 		uint32_t key, uint32_t state);
 
 /**
  * Notify the seat that a key has been pressed on the keyboard. Defers to any
  * keyboard grabs.
  */
-void wlr_seat_keyboard_notify_key(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_keyboard_notify_key(struct wlr_seat *seat, uint32_t time_msec,
 		uint32_t key, uint32_t state);
 
 /**
@@ -517,14 +542,14 @@ struct wlr_touch_point *wlr_seat_touch_get_point(struct wlr_seat *seat,
  * the touch device.
  */
 uint32_t wlr_seat_touch_notify_down(struct wlr_seat *seat,
-		struct wlr_surface *surface, uint32_t time, int32_t touch_id, double sx,
-		double sy);
+		struct wlr_surface *surface, uint32_t time_msec,
+		int32_t touch_id, double sx, double sy);
 
 /**
  * Notify the seat that the touch point given by `touch_id` is up. Defers to any
  * grab of the touch device.
  */
-void wlr_seat_touch_notify_up(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_touch_notify_up(struct wlr_seat *seat, uint32_t time_msec,
 		int32_t touch_id);
 
 /**
@@ -533,7 +558,7 @@ void wlr_seat_touch_notify_up(struct wlr_seat *seat, uint32_t time,
  * even if the surface is not the owner of the touch point for processing by
  * grabs.
  */
-void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time_msec,
 		int32_t touch_id, double sx, double sy);
 
 /**
@@ -542,13 +567,13 @@ void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time,
  * `wlr_seat_touch_point_clear_focus()`.
  */
 void wlr_seat_touch_point_focus(struct wlr_seat *seat,
-		struct wlr_surface *surface, uint32_t time, int32_t touch_id, double sx,
-		double sy);
+		struct wlr_surface *surface, uint32_t time_msec,
+		int32_t touch_id, double sx, double sy);
 
 /**
  * Clear the focused surface for the touch point given by `touch_id`.
  */
-void wlr_seat_touch_point_clear_focus(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_touch_point_clear_focus(struct wlr_seat *seat, uint32_t time_msec,
 		int32_t touch_id);
 
 /**
@@ -560,8 +585,8 @@ void wlr_seat_touch_point_clear_focus(struct wlr_seat *seat, uint32_t time,
  * `wlr_seat_touch_notify_down()` to respect any grabs of the touch device.
  */
 uint32_t wlr_seat_touch_send_down(struct wlr_seat *seat,
-		struct wlr_surface *surface, uint32_t time, int32_t touch_id, double sx,
-		double sy);
+		struct wlr_surface *surface, uint32_t time_msec,
+		int32_t touch_id, double sx, double sy);
 
 /**
  * Send a touch up event for the touch point given by the `touch_id`. The event
@@ -569,7 +594,7 @@ uint32_t wlr_seat_touch_send_down(struct wlr_seat *seat,
  * event. This will remove the touch point. Compositors should use
  * `wlr_seat_touch_notify_up()` to respect any grabs of the touch device.
  */
-void wlr_seat_touch_send_up(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_touch_send_up(struct wlr_seat *seat, uint32_t time_msec,
 		int32_t touch_id);
 
 /**
@@ -578,7 +603,7 @@ void wlr_seat_touch_send_up(struct wlr_seat *seat, uint32_t time,
  * down event. Compositors should use `wlr_seat_touch_notify_motion()` to
  * respect any grabs of the touch device.
  */
-void wlr_seat_touch_send_motion(struct wlr_seat *seat, uint32_t time,
+void wlr_seat_touch_send_motion(struct wlr_seat *seat, uint32_t time_msec,
 		int32_t touch_id, double sx, double sy);
 
 /**
@@ -597,9 +622,51 @@ bool wlr_seat_touch_has_grab(struct wlr_seat *seat);
  */
 bool wlr_seat_validate_grab_serial(struct wlr_seat *seat, uint32_t serial);
 
+/**
+ * Check whether this serial is valid to start a pointer grab action.
+ */
+bool wlr_seat_validate_pointer_grab_serial(struct wlr_seat *seat,
+	struct wlr_surface *origin, uint32_t serial);
+
+/**
+ * Check whether this serial is valid to start a touch grab action. If it's the
+ * case and point_ptr is non-NULL, *point_ptr is set to the touch point matching
+ * the serial.
+ */
+bool wlr_seat_validate_touch_grab_serial(struct wlr_seat *seat,
+	struct wlr_surface *origin, uint32_t serial,
+	struct wlr_touch_point **point_ptr);
+
+/**
+ * Return a new serial (from wl_display_serial_next()) for the client, and
+ * update the seat client's set of valid serials. Use this for all input
+ * events; otherwise wlr_seat_client_validate_event_serial() may fail when
+ * handed a correctly functioning client's request serials.
+ */
+uint32_t wlr_seat_client_next_serial(struct wlr_seat_client *client);
+
+/**
+ * Return true if the serial number could have been produced by
+ * wlr_seat_client_next_serial() and is "older" (by less than UINT32_MAX/2) than
+ * the current display serial value.
+ *
+ * This function should have no false negatives, and the only false positive
+ * responses allowed are for elements that are still "older" than the current
+ * display serial value and also older than all serial values remaining in
+ * the seat client's serial ring buffer, if that buffer is also full.
+ */
+bool wlr_seat_client_validate_event_serial(struct wlr_seat_client *client,
+	uint32_t serial);
+
+/**
+ * Get a seat client from a seat resource. Returns NULL if inert.
+ */
 struct wlr_seat_client *wlr_seat_client_from_resource(
 	struct wl_resource *resource);
 
+/**
+ * Get a seat client from a pointer resource. Returns NULL if inert.
+ */
 struct wlr_seat_client *wlr_seat_client_from_pointer_resource(
 	struct wl_resource *resource);
 
