@@ -247,16 +247,10 @@ static void gles2_wl_drm_buffer_get_size(struct wlr_renderer *wlr_renderer,
 	eglQueryWaylandBufferWL(renderer->egl->display, buffer, EGL_HEIGHT, height);
 }
 
-static int gles2_get_dmabuf_formats(struct wlr_renderer *wlr_renderer,
-		int **formats) {
+static const struct wlr_drm_format_set *gles2_get_dmabuf_formats(
+		struct wlr_renderer *wlr_renderer) {
 	struct wlr_gles2_renderer *renderer = gles2_get_renderer(wlr_renderer);
-	return wlr_egl_get_dmabuf_formats(renderer->egl, formats);
-}
-
-static int gles2_get_dmabuf_modifiers(struct wlr_renderer *wlr_renderer,
-		int format, uint64_t **modifiers) {
-	struct wlr_gles2_renderer *renderer = gles2_get_renderer(wlr_renderer);
-	return wlr_egl_get_dmabuf_modifiers(renderer->egl, format, modifiers);
+	return wlr_egl_get_dmabuf_formats(renderer->egl);
 }
 
 static enum wl_shm_format gles2_preferred_read_format(
@@ -312,7 +306,7 @@ static bool gles2_read_pixels(struct wlr_renderer *wlr_renderer,
 
 	glGetError(); // Clear the error flag
 
-	unsigned char *p = data + dst_y * stride;
+	unsigned char *p = (unsigned char *)data + dst_y * stride;
 	uint32_t pack_stride = width * fmt->bpp / 8;
 	if (pack_stride == stride && dst_x == 0 && flags != NULL) {
 		// Under these particular conditions, we can read the pixels with only
@@ -402,7 +396,6 @@ static const struct wlr_renderer_impl renderer_impl = {
 	.resource_is_wl_drm_buffer = gles2_resource_is_wl_drm_buffer,
 	.wl_drm_buffer_get_size = gles2_wl_drm_buffer_get_size,
 	.get_dmabuf_formats = gles2_get_dmabuf_formats,
-	.get_dmabuf_modifiers = gles2_get_dmabuf_modifiers,
 	.preferred_read_format = gles2_preferred_read_format,
 	.read_pixels = gles2_read_pixels,
 	.texture_from_pixels = gles2_texture_from_pixels,
@@ -552,6 +545,7 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	renderer->exts_str = (const char *)glGetString(GL_EXTENSIONS);
 	wlr_log(WLR_INFO, "Using %s", glGetString(GL_VERSION));
 	wlr_log(WLR_INFO, "GL vendor: %s", glGetString(GL_VENDOR));
+	wlr_log(WLR_INFO, "GL renderer: %s", glGetString(GL_RENDERER));
 	wlr_log(WLR_INFO, "Supported GLES2 extensions: %s", renderer->exts_str);
 
 	if (!check_gl_ext(renderer->exts_str, "GL_EXT_texture_format_BGRA8888")) {
