@@ -78,6 +78,7 @@ const char *atom_map[ATOM_LAST] = {
 	"XdndActionCopy",
 	"XdndActionAsk",
 	"XdndActionPrivate",
+	"_NET_CLIENT_LIST",
 };
 
 static const struct wlr_surface_role xwayland_surface_role;
@@ -212,6 +213,28 @@ static void xwm_send_wm_message(struct wlr_xwayland_surface *surface,
 	xcb_flush(xwm->xcb_conn);
 }
 
+static void xwm_set_net_client_list(struct wlr_xwm *xwm) {
+	size_t mapped_surfaces = 0;
+	struct wlr_xwayland_surface *surface;
+	wl_list_for_each(surface, &xwm->surfaces, link) {
+		if (surface->mapped) {
+			mapped_surfaces++;
+		}
+	}
+
+	xcb_window_t windows[mapped_surfaces + 1];
+	size_t index = 0;
+	wl_list_for_each(surface, &xwm->surfaces, link) {
+		if (surface->mapped) {
+			windows[index++] = surface->window_id;
+		}
+	}
+
+	xcb_change_property(xwm->xcb_conn, XCB_PROP_MODE_REPLACE,
+			xwm->screen->root, xwm->atoms[_NET_CLIENT_LIST],
+			XCB_ATOM_WINDOW, 32, mapped_surfaces, windows);
+}
+
 static void xwm_send_focus_window(struct wlr_xwm *xwm,
 		struct wlr_xwayland_surface *xsurface) {
 	if (!xsurface) {
@@ -268,7 +291,7 @@ static void xwm_surface_activate(struct wlr_xwm *xwm,
 
 static void xsurface_set_net_wm_state(struct wlr_xwayland_surface *xsurface) {
 	struct wlr_xwm *xwm = xsurface->xwm;
-	uint32_t property[3];
+	uint32_t property[4];
 	int i;
 
 	i = 0;
@@ -560,10 +583,22 @@ static void read_surface_normal_hints(struct wlr_xwm *xwm,
 	memcpy(xsurface->size_hints, &size_hints,
 		sizeof(struct wlr_xwayland_surface_size_hints));
 
-	if ((size_hints.flags & XCB_ICCCM_SIZE_HINT_P_MIN_SIZE) == 0) {
+	bool has_min_size_hints = (size_hints.flags & XCB_ICCCM_SIZE_HINT_P_MIN_SIZE) != 0;
+	bool has_base_size_hints = (size_hints.flags & XCB_ICCCM_SIZE_HINT_BASE_SIZE) != 0;
+	/* ICCCM says that if absent, min size is equal to base size and vice versa */
+	if (!has_min_size_hints && !has_base_size_hints) {
 		xsurface->size_hints->min_width = -1;
 		xsurface->size_hints->min_height = -1;
+		xsurface->size_hints->base_width = -1;
+		xsurface->size_hints->base_height = -1;
+	} else if (!has_base_size_hints) {
+		xsurface->size_hints->base_width = xsurface->size_hints->min_width;
+		xsurface->size_hints->base_height = xsurface->size_hints->min_height;
+	} else if (!has_min_size_hints) {
+		xsurface->size_hints->min_width = xsurface->size_hints->base_width;
+		xsurface->size_hints->min_height = xsurface->size_hints->base_height;
 	}
+
 	if ((size_hints.flags & XCB_ICCCM_SIZE_HINT_P_MAX_SIZE) == 0) {
 		xsurface->size_hints->max_width = -1;
 		xsurface->size_hints->max_height = -1;
@@ -702,6 +737,7 @@ static void xwayland_surface_role_commit(struct wlr_surface *wlr_surface) {
 	if (!surface->mapped && wlr_surface_has_buffer(surface->surface)) {
 		wlr_signal_emit_safe(&surface->events.map, surface);
 		surface->mapped = true;
+		xwm_set_net_client_list(surface->xwm);
 	}
 }
 
@@ -718,6 +754,7 @@ static void xwayland_surface_role_precommit(struct wlr_surface *wlr_surface) {
 		if (surface->mapped) {
 			wlr_signal_emit_safe(&surface->events.unmap, surface);
 			surface->mapped = false;
+			xwm_set_net_client_list(surface->xwm);
 		}
 	}
 }
@@ -768,8 +805,9 @@ static void xwm_map_shell_surface(struct wlr_xwm *xwm,
 
 static void xsurface_unmap(struct wlr_xwayland_surface *surface) {
 	if (surface->mapped) {
-		surface->mapped = false;
 		wlr_signal_emit_safe(&surface->events.unmap, surface);
+		surface->mapped = false;
+		xwm_set_net_client_list(surface->xwm);
 	}
 
 	if (surface->surface_id) {
@@ -889,6 +927,12 @@ static void xwm_handle_map_request(struct wlr_xwm *xwm,
 
 	xsurface_set_wm_state(xsurface, ICCCM_NORMAL_STATE);
 	xsurface_set_net_wm_state(xsurface);
+
+	uint32_t values[1];
+	values[0] = XCB_STACK_MODE_BELOW;
+	xcb_configure_window(xwm->xcb_conn, ev->window,
+			XCB_CONFIG_WINDOW_STACK_MODE, values);
+
 	xcb_map_window(xwm->xcb_conn, ev->window);
 }
 
@@ -1706,6 +1750,7 @@ struct wlr_xwm *xwm_create(struct wlr_xwayland *wlr_xwayland) {
 		xwm->atoms[_NET_WM_STATE_FULLSCREEN],
 		xwm->atoms[_NET_WM_STATE_MAXIMIZED_VERT],
 		xwm->atoms[_NET_WM_STATE_MAXIMIZED_HORZ],
+		xwm->atoms[_NET_CLIENT_LIST],
 	};
 	xcb_change_property(xwm->xcb_conn,
 		XCB_PROP_MODE_REPLACE,
